import matplotlib.pyplot as plt
from py_banshee.bn_plot import bn_visualize
from py_banshee.prediction import conditional_margins_hist, inference
from py_banshee.rankcorr import bn_rankcorr


class BansheeCalc:
    def __init__(self, nodeList, edgeList, df):

        self.ParentCell = [None] * (max(nodeList) + 1)
        self.ParentCell = [[] for i in self.ParentCell]
        for v1, v2 in edgeList:
            self.ParentCell[v2].append(v1)

        self.df = df
        self.columns_data = list(df.columns)

        self.df = df
        self.R = None
        self.F = None
        self.calcRankCorr()

    def calcRankCorr(self):
        try:  # можно ли как-то избежать try/except, для случая когда много пропусков и функция не работает
            self.R = bn_rankcorr(
                self.ParentCell,
                self.df,
                var_names=self.columns_data,
                is_data=True,
                plot=False,
            )
        except:
            self.R = None

    def saveGraph(self):
        plt.close()
        plt.cla()
        plt.clf()

        fig_name = "graph"

        columns_with_com = list(map(lambda x: f'"{x}"', self.columns_data))

        bn_visualize(self.ParentCell, self.R, columns_with_com, fig_name=fig_name)

    def getRankCorr(self):
        return self.R

    def getInference(self, len_input_list):
        nodes = list(
            range(len_input_list)
        )  # all variables except for value of interest
        values = self.df.iloc[:, nodes].to_numpy()  # data for predictions
        output = "mean"  # show only mean of the uncertainty distribution
        sampleSize = 10000  # draw 10,000 samples when conditionalizing the BN
        # interp = 'next'  # use the 'next' method to interpolate the empirical
        interp = "linear"
        # interp = 'nearest'

        F = inference(
            Nodes=nodes,
            Values=values,
            R=self.R,
            DATA=self.df,
            Output=output,
            SampleSize=sampleSize,
            Interp=interp,
        )
        self.F = F
        # self.plotHist(len_input_list)
        return F

    def plotHist(self, len_input_list):
        nodes = list(
            range(len_input_list)
        )  # all variables except for value of interest
        values = self.df.iloc[:, nodes].to_numpy()  # data for predictions

        conditional_margins_hist(
            F=self.F, DATA=self.df, names=self.df.columns, condition_nodes=nodes
        )



import networkx as nx
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure
from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtWidgets import QSplitter, QVBoxLayout, QWidget

from bn_modeller.bayesian_nets.graph_preparation import GraphPreparation
from bn_modeller.models import FilterPairTableSQLProxyModel, PairTableSQLProxyModel
from bn_modeller.utils.model_adapters import tablemodel_to_dataframe
from bn_modeller.widgets.extended_slider_widget import ExtendedSliderWidget


class BayesianNetCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=12, height=12, dpi=100):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        super().__init__(self.fig)
        self.bn_ax = self.fig.add_subplot(1, 1, 1)

    def update_plot(self, graph):
        self.bn_ax.clear()
        # self.bn_ax = self.fig.add_subplot(1, 1, 1)

        elarge = [(u, v) for (u, v, d) in graph.edges(data=True) if d["weight"] > 0.5]
        esmall = [(u, v) for (u, v, d) in graph.edges(data=True) if d["weight"] <= 0.5]

        pos = nx.drawing.nx_agraph.graphviz_layout(graph, prog="dot")

        # nodes
        nx.draw_networkx_nodes(graph, pos, node_size=15, ax=self.bn_ax)

        # edges
        nx.draw_networkx_edges(
            graph, pos, edgelist=elarge, width=1, alpha=0.4, ax=self.bn_ax
        )
        nx.draw_networkx_edges(
            graph,
            pos,
            edgelist=esmall,
            width=1,
            alpha=0.4,
            edge_color="b",
            style="dashed",
            ax=self.bn_ax,
        )

        # node labels
        nx.draw_networkx_labels(
            graph,
            pos,
            font_size=12,
            font_family="sans-serif",
            verticalalignment="bottom",
            ax=self.bn_ax,
        )

        # edge weight labels
        edge_labels = nx.get_edge_attributes(graph, "weight")
        nx.draw_networkx_edge_labels(
            graph, pos, edge_labels, font_size=12, ax=self.bn_ax
        )
        self.fig.tight_layout(pad=3.0)
        self.draw()


class BayesianNetView(QSplitter):
    file_path_changed = Signal()

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.depModel: FilterPairTableSQLProxyModel = None
        self._init_ui()

    def _init_ui(self):
        self.setOrientation(Qt.Orientation.Vertical)
        # BN Plot
        bnPlotLayout = QVBoxLayout()

        self.bn_canvas = BayesianNetCanvas()
        self.toolbar = NavigationToolbar2QT(self.bn_canvas, self)
        bnPlotLayout.addWidget(self.toolbar)
        bnPlotLayout.addWidget(self.bn_canvas)
        bnPlotWidget = QWidget()
        bnPlotWidget.setLayout(bnPlotLayout)
        self.addWidget(bnPlotWidget)

        # BN Visualization Settings
        self.threadSliderWidget = ExtendedSliderWidget()
        self.threadSliderWidget.setRange(0, 1)
        self.threadSliderWidget.setValue(0.5)
        self.threadSliderWidget.setTrackMovements(False)
        self.threadSliderWidget.valueChanged.connect(self.drawBN)
        self.addWidget(self.threadSliderWidget)

    def setModels(self, depModel: FilterPairTableSQLProxyModel):
        self.depModel = depModel
        self.depModel.filterInvalidated.connect(self.drawBN)
        self.depModel.dataChanged.connect(self.drawBN)

        self.drawBN()

    @Slot()
    def drawBN(self):
        if self.depModel is None or not self.isVisible():
            return
        print("BayesianNetView.drawBN")
        corr_matrix = tablemodel_to_dataframe(
            self.depModel, role=PairTableSQLProxyModel.PearsonCorrRole
        )
        linkTable = tablemodel_to_dataframe(
            self.depModel, role=Qt.ItemDataRole.CheckStateRole
        )
        graph = GraphPreparation(
            corr_matrix, linkTable, self.threadSliderWidget.value()
        )

        graph.drop_cycle()
        # self.changeLinkTable()

        self.bn_canvas.update_plot(graph.renaming())
        # import pickle
        # pickle.dump(self.graph, open('graph.txt', 'w'))
        # print(self.graph.G.nodes())

        # nx.write_adjlist(self.graph.renaming(), 'graph.txt')

    def showEvent(self, event):
        v = super().showEvent(event)
        self.drawBN()
        return v



import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.special import erfcinv

F = -1 / (2 ** (1 / 2) * erfcinv(3 / 2))


def check_non_select_table(df):
    k = []
    for i in df.columns:
        k.append((df[i] == 0).all())
    return np.all(k)


def find_zero_columns(df):
    """
    Для нахождения столбцов, где все значения пусты
    :param df:
    :return: список столбцов
    """
    return df.loc[:, df.isna().all()].columns


def to_mad_scale(x, m):
    return abs(x - m)


def get_index_outliers(df) -> set:
    """
    Повторяет функцию rmoutliers (median) из MATLAB
    :param df:
    :return: номера индексов выбросов
    """
    outliers_index = []
    for column in df.columns:
        m = df[column].median()
        x = df[column].apply(to_mad_scale, m=m)
        res = F * x.median()

        down = m - res * 3
        up = m + res * 3

        outliers_index.extend(
            list(df.loc[(df[column] >= up) | (df[column] <= down)].index)
        )

    return set(outliers_index)


def get_outliers_cooks(df):
    """
    Нахождение выбросов на основе расстояния Кука
    """
    # storing dependant values
    y = df.iloc[:, 1]

    # storing independent values
    x = df.iloc[:, 0]

    # add bias
    x = sm.add_constant(x)

    # create and train linear regression model
    sm_model = sm.regression.linear_model.OLS(y, x).fit()
    influence = sm_model.get_influence()

    influence_list = influence.cooks_distance[0]
    influence_df = pd.DataFrame(influence_list, columns=["influence"], index=df.index)

    original_length = len(df)

    cooks_df = df.join(influence_df)
    cooks_threshold = 4 / original_length
    cooks_outliers = cooks_df[cooks_df["influence"] > cooks_threshold]

    return set(cooks_outliers.index)



from PySide6.QtCore import QObject
from PySide6.QtSql import (
    QSqlDatabase,
    QSqlQuery,
    QSqlRelation,
    QSqlRelationalTableModel,
)

from bn_modeller.models.feature_sqltable_model import FeatureSqlTableModel


class SampleSqlTableModel(QSqlRelationalTableModel):
    table_name = "sample"
    column_id = "id"
    column_sample_id = "sample_id"
    column_feature_id = "feature_id"
    column_value = "value"

    def __init__(self, parent: QObject = None, db: QSqlDatabase = None):
        super().__init__(parent, db)

        self.setRelation(
            2,
            QSqlRelation(
                aTableName=FeatureSqlTableModel.table_name,
                indexCol=SampleSqlTableModel.column_feature_id,
                displayCol=FeatureSqlTableModel.column_name,
            ),
        )
        self.setJoinMode(QSqlRelationalTableModel.JoinMode.LeftJoin)

        query = QSqlQuery(
            f"CREATE TABLE IF NOT EXISTS {SampleSqlTableModel.table_name} (\
                          {SampleSqlTableModel.column_id} INTEGER PRIMARY KEY AUTOINCREMENT, \
                          {SampleSqlTableModel.column_sample_id} INTEGER, \
                          {SampleSqlTableModel.column_feature_id} INTEGER, \
                          {SampleSqlTableModel.column_value} REAL,\
                          FOREIGN KEY({SampleSqlTableModel.column_feature_id}) REFERENCES {FeatureSqlTableModel.table_name}({FeatureSqlTableModel.column_id})\
                          );\
                          "
        )

        if not query.exec():
            raise RuntimeError(f"Unable to connect to DB: {query.lastError()}")
        self.setTable(SampleSqlTableModel.table_name)

    def insertRowIntoTable(self, values):
        return super().insertRowIntoTable(values)



import pandas as pd
import scipy.stats as stats
import seaborn as sns
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from PySide6 import QtCore, QtWidgets

from bn_modeller.bayesian_nets.utils import get_index_outliers, get_outliers_cooks


class JointPlot(QtWidgets.QMainWindow):

    def __init__(
        self,
        parent=None,
        x=None,
        y=None,
        x_outlier=None,
        y_outlier=None,
        df_without_nan=None,
        column_name1=None,
        column_name2=None,
        r=None,
        r_w=None,
    ):
        super(JointPlot, self).__init__(parent)

        # self.setWindowTitle('Матрица')

        self.main_widget = QtWidgets.QWidget(self)

        self.resize(600, 600)

        self.fig = Figure()

        self.fig = self.initFigure(
            x,
            y,
            x_outlier,
            y_outlier,
            df_without_nan,
            column_name1,
            column_name2,
            r,
            r_w,
        )

        self.canvas = FigureCanvas(self.fig)

        self.canvas.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.canvas.updateGeometry()
        # self.label = QtWidgets.QLabel("A plot:")
        toolbar = NavigationToolbar(self.canvas, self)

        self.layout = QtWidgets.QGridLayout(self.main_widget)
        self.layout.addWidget(toolbar)
        # self.layout.addWidget(self.label)
        self.layout.addWidget(self.canvas)

        self.setCentralWidget(self.main_widget)
        self.show()
        self.fig.tight_layout()

    def initFigure(
        self,
        x,
        y,
        x_outlier,
        y_outlier,
        df_without_nan,
        column_name1,
        column_name2,
        r,
        r_w,
    ):

        g = sns.JointGrid()
        # g.fig.set_size_inches((8, 8))
        g.ax_marg_x.set_xlim(left=min(x) - 1, right=max(x) + 1)

        sns.scatterplot(
            x=x, y=y, s=50, linewidth=0, ax=g.ax_joint, color="tab:blue", alpha=0.7
        )
        sns.scatterplot(
            x=x_outlier,
            y=y_outlier,
            s=50,
            linewidth=0,
            ax=g.ax_joint,
            color="tab:red",
            alpha=0.7,
        )

        sns.regplot(
            x=df_without_nan[column_name1],
            y=df_without_nan[column_name2],
            ax=g.ax_joint,
            color="r",
            scatter=False,
            line_kws={"linewidth": 1},
        )
        sns.regplot(x=x, y=y, ax=g.ax_joint, color="black", scatter=False)

        sns.kdeplot(y=y, linewidth=2, ax=g.ax_marg_y)
        sns.kdeplot(x=x, linewidth=2, ax=g.ax_marg_x)
        g.ax_joint.annotate(
            f"$R = {r:.3f}$ - все точки",
            xy=(0.1, 0.9),
            xycoords="axes fraction",
            ha="left",
            va="center",
            color="tab:red",
        )

        g.ax_joint.annotate(
            f"$R = {r_w:.3f}$ - без выбросов",
            xy=(0.1, 0.85),
            xycoords="axes fraction",
            ha="left",
            va="center",
            color="tab:blue",
        )
        return g.fig


class jointgridWindow(QtWidgets.QWidget):
    def __init__(self, parent=None, data=None, predicted_data=pd.DataFrame()):
        super(jointgridWindow, self).__init__(parent)

        self.error_dialog = QtWidgets.QErrorMessage()

        self.df = data.join(predicted_data)
        col_name = self.df.columns

        self.df_select = pd.DataFrame(columns=["select"], index=col_name)
        self.tableWidget = QtWidgets.QTableWidget(len(col_name), 1)
        self.build_button = QtWidgets.QPushButton("Plot")

        # self.canvas = FigureCanvas(self.g.fig)
        # self.g = self.fig.add_subplot(111)

        plot_layout = QtWidgets.QVBoxLayout()
        # plot_layout.addWidget(self.canvas)

        self.select_layout = QtWidgets.QVBoxLayout(self)
        self.select_layout.addWidget(self.tableWidget)
        # self.select_layout.addWidget(self.build_button)

        main_layout = QtWidgets.QHBoxLayout()
        main_layout.addLayout(plot_layout)
        main_layout.addLayout(self.select_layout)

        self.selected_items = []
        self.selected_items_names = []
        self.num_of_selected_items = 2

        self.tableWidget.setHorizontalHeaderLabels(["select"])
        self.tableWidget.setVerticalHeaderLabels(col_name)
        for i in range(self.tableWidget.rowCount()):
            for j in range(2):
                item = QtWidgets.QTableWidgetItem()
                item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
                item.setCheckState(QtCore.Qt.Unchecked)
                # item.setCheckState(self.state[i][j])
                self.tableWidget.setItem(i, j, item)

        self.tableWidget.cellChanged.connect(self.turnOffClicked)
        self.build_button.clicked.connect(self.on_pushButton_clicked)
        #
        # self.show()
        self.dialogs = list()

    def turnOffClicked(self, row, column):
        item = self.tableWidget.item(row, column)
        item_name = self.tableWidget.verticalHeaderItem(row).text()
        if item.checkState() == QtCore.Qt.Checked:
            self.selected_items.append(item)
            self.selected_items_names.append(item_name)
            if len(self.selected_items) > self.num_of_selected_items:
                for item_ in self.selected_items[: -self.num_of_selected_items]:
                    self.selected_items.pop(0)
                    self.selected_items_names.pop(0)
                    item_.setCheckState(QtCore.Qt.Unchecked)
        elif item.checkState() == QtCore.Qt.Unchecked and item in self.selected_items:
            delete_index = self.selected_items.index(item)
            self.selected_items.pop(delete_index)
            self.selected_items_names.pop(delete_index)

    def updateData(self, df_predicted):

        if not set(df_predicted.columns).intersection(set(self.df.columns)):
            self.df = self.df.join(df_predicted)

        col_name = self.df.columns

        self.df_select = pd.DataFrame(columns=["select"], index=col_name)
        self.tableWidget = QtWidgets.QTableWidget(len(col_name), 1)
        # self.build_button = QtWidgets.QPushButton("Построить")
        #
        # plot_layout = QtWidgets.QVBoxLayout()
        # # plot_layout.addWidget(self.canvas)
        #
        # select_layout = QtWidgets.QVBoxLayout(self)
        self.select_layout.addWidget(self.tableWidget)
        self.select_layout.addWidget(self.build_button)
        #
        # main_layout = QtWidgets.QHBoxLayout()
        # main_layout.addLayout(plot_layout)
        # main_layout.addLayout(select_layout)
        #
        # self.selected_items = []
        # self.selected_items_names = []
        # self.num_of_selected_items = 2

        self.tableWidget.setHorizontalHeaderLabels(["select"])
        self.tableWidget.setVerticalHeaderLabels(col_name)
        for i in range(self.tableWidget.rowCount()):
            for j in range(2):
                item = QtWidgets.QTableWidgetItem()
                item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
                item.setCheckState(QtCore.Qt.Unchecked)
                # item.setCheckState(self.state[i][j])
                self.tableWidget.setItem(i, j, item)

        self.tableWidget.cellChanged.connect(self.turnOffClicked)
        # self.build_button.clicked.connect(self.on_pushButton_clicked)

        # self.show()

    def removeTableWidget(self):
        self.tableWidget.deleteLater()

    def on_pushButton_clicked(self):

        if len(self.selected_items) == self.num_of_selected_items:
            column_name1 = self.selected_items_names[0]
            column_name2 = self.selected_items_names[1]
            df_without_nan = self.df.dropna(subset=[column_name1, column_name2])

            if len(df_without_nan) < 2:
                self.error_dialog.showMessage("Not enough intersecting data")
                return 1

            # нахождение выбросов
            # outliers_index = get_index_outliers(df_without_nan[[column_name1, column_name2]])
            outliers_index = get_outliers_cooks(
                df_without_nan[[column_name1, column_name2]]
            )

            # датафрейм выбросов и без выбросов
            df_outliers = df_without_nan.loc[list(outliers_index)]
            df_without_outliers = df_without_nan.drop(outliers_index)

            # корреляция полного датафрейма и без выбросов
            r, _ = stats.pearsonr(
                df_without_nan[column_name1], df_without_nan[column_name2]
            )
            r_w, _ = stats.pearsonr(
                df_without_outliers[column_name1], df_without_outliers[column_name2]
            )

            # точки где нет выбросов, точки с выбросами
            x, y = df_without_outliers[column_name1], df_without_outliers[column_name2]
            x_outlier, y_outlier = df_outliers[column_name1], df_outliers[column_name2]

            dialog = JointPlot(
                self,
                x=x,
                y=y,
                x_outlier=x_outlier,
                y_outlier=y_outlier,
                df_without_nan=df_without_nan,
                column_name1=column_name1,
                column_name2=column_name2,
                r=r,
                r_w=r_w,
            )
            self.dialogs.append(dialog)
            dialog.show()



import os

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QAction, QGuiApplication
from PySide6.QtSql import QSqlDatabase, QSqlQuery
from PySide6.QtWidgets import QMenuBar, QStackedWidget, QStyle, QTabWidget, QWidget

from bn_modeller.dialogs import AboutDialog
from bn_modeller.models import DependencyManyToManySqlTableModel, PairTableSQLProxyModel
from bn_modeller.models.feature_sqltable_model import FeatureSqlTableModel
from bn_modeller.models.sample_sqltable_model import SampleSqlTableModel
from bn_modeller.widgets.page.bayesian_network_page import BayesianNetworkPageWidget
from bn_modeller.widgets.page.database_page import DatabasePageWidget
from bn_modeller.widgets.project_wizard import ProjectLoadWizard
from bn_modeller.windows.base_window import BaseWindow


class MainWindow(BaseWindow):
    go_back = Signal()

    def __init__(self, parent: QWidget | None = None, flags=Qt.WindowType()):

        super().__init__("", parent, flags)

        self._main_widget: QTabWidget

        self._title = self.tr("Bayesian Network Modeller")

        self._init_ui()

        self._views_history: list[QWidget] = []
        self._project_path = None
        self._db: QSqlDatabase = None

        self.featureSqlTableModel: FeatureSqlTableModel = None
        self.sampleSqlTableModel: SampleSqlTableModel = None
        self.projectWizard: ProjectLoadWizard = None

        QGuiApplication.instance().applicationStateChanged.connect(
            self.application_state_changed
        )

    def _init_db(self):
        self.featureSqlTableModel = FeatureSqlTableModel(db=self._db)
        self.sampleSqlTableModel = SampleSqlTableModel(db=self._db)
        self._initCacheDbInMemory()

        self._dependencyManyToManySqlTableModel = DependencyManyToManySqlTableModel(
            db=self._db
        )
        self._dependPairModel = PairTableSQLProxyModel(
            self.featureSqlTableModel, db=self._db
        )

        self.databasePageWidget.setModels(
            featureSqlTableModel=self.featureSqlTableModel,
            sampleSqlTableModel=self.sampleSqlTableModel,
        )

        self.dependencySetupPageWidget.setModels(
            pairTableSQLProxyModel=self._dependPairModel
        )

    def _init_ui(self):
        self._main_widget = QTabWidget()
        self.set_central_title(self._title)

        self.databasePageWidget = DatabasePageWidget()
        self._main_widget.addTab(self.databasePageWidget, self.tr("Database"))

        self.dependencySetupPageWidget = BayesianNetworkPageWidget()
        self._main_widget.addTab(
            self.dependencySetupPageWidget, self.tr("Bayesian Networks")
        )

        self.setCentralWidget(self._main_widget)

        # add_data_action = QAction(self.style().standardIcon(
        #     QStyle.StandardPixmap.SP_FileDialogContentsView), '&AddData', self)
        # add_data_action.setStatusTip(self.tr('Add Data'))
        # add_data_action.triggered.connect(self.add_data_clicked)
        # self.getMainToolBar().addAction(add_data_action)

        # Create a menu bar
        self._menu_bar = QMenuBar()
        self.setMenuBar(self._menu_bar)

        # Create the "Help" menu
        self._help_menu = self._menu_bar.addMenu(self.tr("Help"))

        # Create an action for the "About" item in the "Help" menu
        about_action = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MessageBoxInformation),
            "&About",
            self,
        )
        about_action.setStatusTip(self.tr("Show About Dialog"))
        about_action.triggered.connect(self.show_about_dialog)

        # Add the action to the "Help" menu
        self._help_menu.addAction(about_action)

    def show_about_dialog(self):
        dialog = AboutDialog()
        dialog.exec()

    def _save_to_history(self, previousWidget: QWidget):
        self._viewsHistory.append(previousWidget)

    def _set_current_widget(self, newCurrentWidget: QWidget):
        self._save_to_history(self._main_widget.currentWidget())
        self._main_widget.setCurrentWidget(newCurrentWidget)

    def _initCacheDbInMemory(self):
        self.featureSqlTableModel.select()
        while self.featureSqlTableModel.canFetchMore():
            self.featureSqlTableModel.fetchMore()
        self.sampleSqlTableModel.select()
        while self.sampleSqlTableModel.canFetchMore():
            self.sampleSqlTableModel.fetchMore()

    @Slot()
    def go_back_clicked(self):
        if len(self._views_history) > 0:
            previousWidget: QWidget = self._views_history.pop()
            self._main_widget.setCurrentWidget(previousWidget)
            self.setCentralTitle("", "")
            self.go_back.emit()

    @Slot()
    def home_clicked(self):
        self._views_history.clear()
        self._main_widget.setCurrentWidget(self._homepageWidget)
        self.setCentralTitle("", "")
        self.go_back.emit()

    @Slot(Qt.ApplicationState)
    def application_state_changed(self, state: Qt.ApplicationState):
        if self._project_path is None and self.projectWizard is None:
            self.projectWizard = ProjectLoadWizard()
            wizard_ret = self.projectWizard.exec()
            if wizard_ret != 1:
                self.close_app()
            self._project_path = self.projectWizard.get_project_path()
            self._db = self.projectWizard._db
            self._init_db()
